"""A module that provides InvalidSchemaError

A module that provides the schema error class ``InvalidSchemaError`` to __init__.
"""
import os
from typing import List, Iterable, Union, Iterator

from jsonschema import ValidationError


_INDENT_UNIT_LENGTH = 4


def __path_part_to_str(path_part: Union[str, int]) -> str:
    """Convert one JSON path parcel to a string.

    Args:
        path_part: A string or integer representing a JSON path compartment.

    Returns:
        A parcel of string representing a JSON path.
    """
    return f"[{path_part}]" \
        if isinstance(path_part, int) \
        else f"[\"{path_part}\"]"


def _path_to_str(path: Iterable[Union[str, int]]) -> str:
    """Convert a JSON path to string.

    Args:
        path: An Iterable representing a JSON path

    Returns:
        A string representing the JSON path
    """
    return "".join(map(__path_part_to_str, path))


class InvalidSchemaError(Exception):
    """Schema error class.

    This error raises if the schema is invalid.

    When the schema is valid but no value satisfy it (exp: ``schema.minimum > schema.maximum``), ``SchemaConflictError``
    is raised.
    """

    def __init__(self,
                 validation_error_list: List[ValidationError]):
        """Initialize InvalidSchemaError

        Args:
            validation_error_list:
                A list of validation errors that caused this error.
        """
        # (List[ValidationErrorWrapper]) A list of validation errors that caused this error.
        self.__cause_list = map(lambda e: ValidationErrorWrapper(e), validation_error_list)

        message_line_list = self._make_message_line_list()

        super().__init__(os.linesep + os.linesep.join(message_line_list))

    def _make_message_line_list(self) -> List[str]:
        """Create an error message

        Returns:
            A list consisting of lines of messages.
        """
        return sum(map(lambda c: c.make_message(), self.__cause_list), [])


class ValidationErrorWrapper:
    """Wrapper of the ``ValidationError`` Class
    """

    @property
    def base_error(self) -> ValidationError:
        """The ValidationError
        """
        return self.__base_error

    @property
    def context(self) -> Iterator:
        """The context of error.
        """
        return map(lambda e: ValidationErrorWrapper(e), self.base_error.context)

    def __init__(self, base_error: ValidationError):
        self.__base_error = base_error

    def make_message(self, indent: int = _INDENT_UNIT_LENGTH) -> List[str]:
        """Generates validation error messages.

        Args:
            indent: The number of spaces at the beginning of each line.

        Returns:
            A list consisting of lines of messages.
        """
        path_str = _path_to_str(self.base_error.absolute_path)
        context_line_list = self._make_message_of_context(indent)

        return [
                   f"{indent * ' '}On instance{path_str}:",
                   f"{(indent + _INDENT_UNIT_LENGTH) * ' '}{self.base_error.message}",
               ] + context_line_list

    def _make_message_of_context(self, indent: int):
        """Create error messages of error context.

        Create a message from all validation errors that caused this.

        Args:
            indent: The number of spaces at the beginning of each line.

        Returns:
            A list consisting of lines of messages.
        """

        def context_to_message_line(e: ValidationErrorWrapper):
            return e.make_message(indent + 2 * _INDENT_UNIT_LENGTH)

        return sum(map(context_to_message_line, self.context), [])



"""Parsers for argparse.ArgumentParser.add_argument
"""


def positive_integer(string: str) -> int:
    value = int(string)

    if value <= 0:
        raise ValueError('invalid literal for positive_int(): ' + repr(string))

    return int(string)



"""
This package provides errors raised in ranjg modules.
"""


from ._options_file_io_error import OptionsFileIOError
from ._schema_file_io_error import SchemaFileIOError
from ._invalid_schema_error import InvalidSchemaError
from ._schema_confrict_error import SchemaConflictError
from ._generate_error import GenerateError, GenerateConflictError



from collections import namedtuple
from typing import Union, Tuple, Optional


class NumberRange(namedtuple("NumberRange", "minimum maximum exclusive_minimum exclusive_maximum")):

    def __contains__(self, item):
        if item < self.minimum:
            return False
        elif item > self.maximum:
            return False
        elif self.exclusive_minimum and item <= self.minimum:
            return False
        elif self.exclusive_maximum and item >= self.maximum:
            return False
        else:
            return True

    def replace(self, **kwargs):
        return self._replace(**kwargs)

    @classmethod
    def from_schema(cls, schema: dict = None):
        """Initialize NumberRange

        Args:
            schema: Schema object.
        """
        if schema is None:
            schema = {}
        return _from_schema(schema)


def _normalize_minimum(minimum: Union[float, int, None],
                       exclusive_minimum: Union[float, int, bool, None]) -> Tuple[Optional[float], bool]:
    """Normalize minimum

    Normalize minimum value of draft-4-style schema and draft-7-style schema to draft-4-style schema.

    Args:
        minimum: The lower bound.
        exclusive_minimum:
            The lower bound or a boolean value meaning that whether or not to include the value ``minimum`` in the
            range.

    Returns:
        The first value is minimum. The second value is whether or not to include the value ``minimum`` in the range.
    """
    if exclusive_minimum is None:
        # exclusive_minimum の指定が無いなら、フラグは立たない。
        exclusive_minimum = False
    elif exclusive_minimum is True:
        # exclusive_minimum が True なら、フラグが立つ。ただし minimum の指定が無い場合はフラグを立てない決まり。
        exclusive_minimum = minimum is not None
    elif exclusive_minimum is False:
        # exclusive_minimum が False なら、フラグが立たない。
        pass
    else:
        # exclusive_minimum が数値なら、minimum 以上なら使用される。
        if minimum is None or minimum <= exclusive_minimum:
            minimum = exclusive_minimum
            exclusive_minimum = True
        else:
            exclusive_minimum = False

    return minimum, exclusive_minimum


def _normalize_maximum(maximum: Union[float, int, None],
                       exclusive_maximum: Union[float, int, bool, None]) -> Tuple[Optional[float], bool]:
    """Normalize minimum

    Normalize minimum value of draft-4-style schema and draft-7-style schema to draft-4-style schema.

    Args:
        maximum: The upper bound.
        exclusive_maximum:
            The upper bound or a boolean value meaning that whether or not to include the value ``maximum`` in the
            range.

    Returns:
        The first value is maximum. The second value is whether or not to include the value ``maximum`` in the range.
    """
    if exclusive_maximum is None:
        # exclusive_maximum の指定が無いなら、フラグは立たない。
        exclusive_maximum = False
    elif exclusive_maximum is True:
        # exclusive_maximum が True なら、フラグが立つ。ただし maximum の指定が無い場合はフラグを立てない決まり。
        exclusive_maximum = maximum is not None
    elif exclusive_maximum is False:
        # exclusive_maximum が False なら、フラグが立たない。
        pass
    else:
        # exclusive_maximum が数値なら、maximum 以下なら使用される。
        if maximum is None or maximum >= exclusive_maximum:
            maximum = exclusive_maximum
            exclusive_maximum = True
        else:
            exclusive_maximum = False

    return maximum, exclusive_maximum


def _from_schema(schema: dict) -> NumberRange:
    """Initialize NumberRange

    Args:
        schema: Schema object.
    """
    minimum, exclusive_minimum = _normalize_minimum(minimum=schema.get("minimum"),
                                                    exclusive_minimum=schema.get("exclusiveMinimum"))
    maximum, exclusive_maximum = _normalize_maximum(maximum=schema.get("maximum"),
                                                    exclusive_maximum=schema.get("exclusiveMaximum"))

    return NumberRange(minimum=minimum,
                       maximum=maximum,
                       exclusive_minimum=exclusive_minimum,
                       exclusive_maximum=exclusive_maximum)



import json

from ..error import SchemaFileIOError


def load(filepath: str) -> dict:
    """load a schema file

    Raises:
        SchemaFileIOError:
            When loading file is failed
    """
    try:
        with open(filepath) as fp:
            loaded_schema = json.load(fp)
    except json.decoder.JSONDecodeError as e:
        raise SchemaFileIOError(f'This file cannot be parsed as schema: {filepath}') from e

    return loaded_schema



import string
import sys
import argparse
from typing import Tuple, Optional, TextIO, Iterable

from .util.listutil import count
from ._arg_parser import positive_integer
from . import gen


__formatter = string.Formatter()


def main():
    """A function to be called when the module is executed on the command line.
    """
    # 引数を取得
    args = parse_args()

    # 出力先を決定
    output_file, output_file_list, output_fp = get_output_target(args)

    gen(schema_file=args.schema_file_path, output_file=output_file, output_fp=output_fp, options_file=args.options,
        multiplicity=args.multiplicity, output_file_list=output_file_list,
        # To ensure that generated value is exposed to garbage collection earlier.
        return_none=True)


def _count_placeholder(format_string: str) -> int:
    return count(lambda t: t[1] is not None, __formatter.parse(format_string))


def parse_args():
    """Get and parse command line arguments.

    Returns:
        An object that holds the values obtained from the command line arguments.
    """
    parser = argparse.ArgumentParser(description="Generate json file randomly according to json schema.")

    parser.add_argument("schema_file_path", help="Path of json schema file. This file is used as base schema.")
    parser.add_argument("--json_output", "-j", help="Path to which json file is written.")
    parser.add_argument("-n", dest="file_num", type=positive_integer,
                        help="The number of output json file. "
                             "Repeats the generation for the specified number of times and outputs the result to each "
                             "file. If it is specified, --json_output is required and used as format of filepath.")
    parser.add_argument("--options", help="Path of options file.")
    parser.add_argument("--list", "-l", dest="multiplicity", type=positive_integer,
                        help="If specified, repeats the generation for the specified number of times "
                             "and outputs the results as a list.")

    args = parser.parse_args()

    if args.file_num is not None:
        # -n の指定時は --json_output が必須
        if args.json_output is None:
            parser.error("the following arguments are required when -n is specified: --json_output")

        # -n に指定がある場合、--json_output は1つのプレースホルダーを持つフォーマットでなくてはならない
        if _count_placeholder(args.json_output) != 1:
            parser.error("when -n is specified, --json_output must have exactly one placeholder such as '{}'")

    return args


def get_output_target(args) -> Tuple[Optional[str], Optional[Iterable[str]], Optional[TextIO]]:
    """Returns the values to be used as ``(output_file, output_file_list, output_fp)``, arguments of ``ranjg.gen``.

    Args:
        args: An object that holds the values obtained from the command line arguments.

    Returns:
        The values to be used as ``(output_file, output_file_list, output_fp)``, arguments of ``ranjg.gen``.
    """
    if args.file_num is not None:
        return None, map(lambda i: args.json_output.format(i), range(args.file_num)), None
    elif args.json_output is not None:
        return args.json_output, None, None
    else:
        return None, None, sys.stdout


if __name__ == "__main__":
    main()



from typing import Tuple, Union, Iterable


class GenerationContext:
    """Context of randomly construction.
    """
    _key_path: Tuple[Union[int, str]]
    _current_schema: dict

    @classmethod
    def root(cls, current_schema: dict):
        return GenerationContext(path=tuple(), current_schema=current_schema)

    def __init__(self, path: Iterable[Union[int, str]], current_schema: dict):
        self._key_path = tuple(path)
        self._current_schema = current_schema

    @property
    def key_path(self) -> Tuple[Union[int, str]]:
        return self._key_path

    def resolve(self, key: Union[int, str], current_schema: dict):
        return GenerationContext(path=(*self._key_path, key), current_schema=current_schema)


class SchemaContext:
    """Context of factory construction.
    """
    _key_path: Tuple[Union[int, str]]
    _current_schema: dict
    _is_for_options: bool = False

    @classmethod
    def root(cls, current_schema: dict):
        return SchemaContext(path=tuple(), current_schema=current_schema)

    @classmethod
    def for_options(cls, current_schema: dict, path: Iterable[Union[int, str]]):
        # Options 内のスキーマを使用する場合の context
        return SchemaContext(path=path, current_schema=current_schema, for_options=True)

    def __init__(self, path: Iterable[Union[int, str]], current_schema: dict, for_options: bool = False):
        self._key_path = tuple(path) if path is not None else None
        self._current_schema = current_schema
        self._is_for_options = for_options

    @property
    def key_path(self) -> Tuple[Union[int, str]]:
        return self._key_path

    def resolve(self, key: Union[int, str], current_schema: dict):
        return SchemaContext(path=(*self._key_path, key), current_schema=current_schema)



from ranjg._context import SchemaContext


class SchemaConflictError(Exception):
    """Conflict errors in the schema.

    This error raises if the schema has conflict.
    In other words, when no value satisfy the schema, this error is raised.

    When the schema is invalid (exp: ``schema.type`` is illegal string), ``InvalidSchemaError`` is raised.
    """
    def __init__(self, message: str, context: SchemaContext):
        super(Exception, self).__init__(message)
        self.context = context



import json
from functools import lru_cache
from typing import NamedTuple, Union

from .error import OptionsFileIOError


NO_COPY = 'NO_COPY'
SHALLOW_COPY = 'SHALLOW_COPY'
DEEP_COPY = 'DEEP_COPY'


class Options(NamedTuple):
    """Options of ``ranjg.gen``.

    See Also:
        :doc:`ranjg-options`:
            To know what options are used in each generation.
    """

    #: The maximum number of generation when multiple attempts are required to obtain a result that satisfies the
    #: schema.
    #: If it is None, regeneration is repeated until a result that matches the schema is obtained.
    #: If it is too small, there is a high probability that the generation will fail and raise an GenerateError.
    regeneration_attempt_limit: int = 50

    #: In bool generation, returns True with probability x, False with probability 1-x.
    default_prob_of_true_given_bool: Union[int, float] = 0.5

    #: In string generation, it is used as the difference between minLength and maxLength, if necessary.
    #: If only one of minLength and maxLength is specified, it is used to determine the other.
    default_length_range_of_genstr: int = 0

    #: In string generation, it is used as the value of minLength when both minLength and maxLength are not specified.
    default_min_length_of_string: int = 1

    #: In string generation, it is used as the value of maxLength when both minLength and maxLength are not specified.
    default_max_length_of_string: int = 10

    #: In dict generation, every optional property in the schema is contained in the result dict with a x probability
    #: independently
    default_prob_of_optional_properties: Union[int, float] = 0.5

    #: In dict generation, it is used to generate properties for which no schema is specified.
    #: For example, a property is required but its schema is not specified.
    default_schema_of_properties: dict = {"type": "null"}

    #: In dict generation, when a property whose name matches each key is generated, the corresponding value is used as
    #: the schema.
    priority_schema_of_properties: dict = {}

    #: In list generation, it is used to generate elements for which no schema is specified.
    #: For example, when `item` is specified as tuple format, and minLength is greater than its length.
    default_schema_of_items: dict = {"type": "null"}

    #: In generation from enum, it specifies how to generate an output value from elements in schema.enum.
    #: For example, if schema.enum contains objects of type list or dict as its content,
    #: and the result value is returned as it is,
    #: any change operation on the result value will change the way schema works.
    #:
    #: If it is options.NO_COPY, generated value has same id for one of element in schema.enum.
    #: If it is options.SHALLOW_COPY, generated value is shallow copy of one of element in schema.enum.
    #: If it is options.DEEP_COPY, generated value is deep copy of one of element in schema.enum.
    #: See the description of copy pickle for information on shallow copy and deep copy.
    enum_copy_style: str = DEEP_COPY

    @classmethod
    @lru_cache(maxsize=1)
    def default(cls):
        """Default options

        It is used when an user don't specify options.

        Returns:
            Options: Default options
        """
        return Options()


def __object_hook_on_load(d: dict) -> Union[Options, dict]:
    # Options を json.load メソッドでロードする際の object_hook として使用する

    # object_hook はルートオブジェクト以外にも使用されるため、たとえば d が
    # default_schema_of_properties に対応する dict である場合もある。
    # そのため、Options にできない場合はルートオブジェクトではないと判断して d をそのまま返す。
    try:
        return Options(**d)
    except TypeError:
        return d


def load(filepath: str) -> Options:
    """load an option file

    Raises:
        OptionsFileIOError:
            When loading file is failed
    """
    try:
        with open(filepath, mode='r') as f:
            # object_hook により、json.load の戻り値は Options にパース可能なら Options に、
            # そうでないなら dict になる。
            options: Union[Options, dict] = json.load(f, object_hook=__object_hook_on_load)
    except json.decoder.JSONDecodeError as e:
        raise OptionsFileIOError(f'This file cannot be parsed as options: {filepath}') from e

    if isinstance(options, Options):
        return options
    else:
        raise OptionsFileIOError(f'This file cannot be parsed as options: {filepath}')



import itertools
import json
from typing import Optional, TextIO, Iterable

import ranjg
from . import schemas
from ._context import GenerationContext
from .options import Options
from .options import load as load_options
from .util.numutil import is_integer


def __count_not_null(*args) -> int:
    return len(tuple(filter(lambda v: v is not None, args)))


def gen(schema: dict = None,
        *,
        schema_file: str = None,
        output_file: str = None,
        output_file_list: Iterable[str] = None,
        output_fp: TextIO = None,
        output_fp_list: Iterable[TextIO] = None,
        options: Optional[Options] = None,
        options_file: str = None,
        multiplicity: Optional[int] = None,
        schema_is_validated: bool = False,
        return_none: bool = False,
        context: Optional[GenerationContext] = None):
    """Generate something randomly according to the JSON schema.

    This function is not fully compliant with the JSON schema, and unsupported parameters in the schema are ignored.
    See also :doc:`ranjg-json-schema` to explore the supported parameters.

    Examples
        The following code is most simple usage.

        >>> import ranjg
        >>> schema_dict = { 'type': 'string' }
        >>> ranjg.gen(schema_dict)    # -> returns some string

        By replacing the contents of ``schema_dict``, you can change the value to be generated.

        If you want to specify a schema without a python dict but with JSON file, you can use argument ``schema_file``.

        >>> import ranjg
        >>> schema_path = './schema_file.json'
        >>> ranjg.gen(schema_file=schema_path)  # -> returns some value

        If you want to get the result as JSON file, you can use argument ``output_file`` or ``output_fp``. (Following
        two examples will run similarly.

        >>> import ranjg
        >>> schema_dict = { 'type': 'string' }
        >>> ranjg.gen(schema_dict, output_file='./output.json')
        >>> # -> returns the result value and writes the result to specified file

        >>> import ranjg
        >>> schema_dict = { 'type': 'string' }
        >>> with open('./output.json', 'w+') as out_fp:
        >>>     ranjg.gen(schema_dict, output_fp=out_fp)
        >>>     # -> returns the result value and writes the result to specified file

    Args:
        schema (dict, optional):
            JSON schema object. See also :doc:`ranjg-json-schema`.
            Only one of this argument or ``schema_file`` needs to be specified.
        schema_file (str, optional):
            The path to JSON schema file. This JSON schema is used instead of the argument ``schema``.
        output_file (str, optional):
            The path to a file where the result will be output as JSON.
            If ``multiplicity`` is specified, a list consisting of the generated values will be output as json.
        output_file_list (Iterable[str], optional):
            The list of paths to a file where the result will be output as JSON.
            It repeats the generation and outputs each result to each file.
        output_fp (TextIO, optional):
            The writing object of a file where the result will be output as JSON.
            If ``multiplicity`` is specified, a list consisting of the generated values will be output as json.
        output_fp_list (Iterable[TextIO], optional):
            The list of writing objects of files where the result will be output as JSON.
            It repeats the generation and outputs each result to each file.
        options (Options, optional):
            The options for generation.
        options_file (str, optional):
            The path to options file. This is parsed as JSON to an Options instance.
        multiplicity (int, optional):
            If specified, it repeats the generation for the specified number of times and returns the results as a list.
        schema_is_validated (bool, optional):
            Whether the schema is already validated or not.
            (In normal usage, this argument is not specified.)
        return_none (bool, default=False):
            If it is True, this function returns None.
            (If it is False, the result contains all outputs.
            In particular, if it is repeated a lot, such as when a large list is specified in output_file_list,
            it may overwhelm memory.)
        context (GenerationContext):
            The context of construction.
            (In normal usage, this argument is not specified. This argument is for using this function recursively.)

    Returns:
        Generated something. It is satisfies the JSON schema.
        However, if ``multiplicity`` is specified, it returns a list, each element of which satisfies the schema.

    Raises:
        SchemaFileIOError:
            When loading schema_file is failed
        OptionsFileIOError:
            When loading options_file is failed
        InvalidSchemaError:
            When the schema specified as arguments is invalid.
        SchemaConflictError:
            When the schema specified as arguments has confliction.
            In other words, when no value can satisfy the schema.
        GenerateError:
            If an unforeseen error arises.
    """
    if schema is None and schema_file is None:
        raise ValueError("schema or schema_file must be specified.")
    if schema is not None and schema_file is not None:
        raise ValueError("Only one of schema and schema_file can be set.")
    if __count_not_null(output_file, output_fp, output_file_list, output_fp_list) >= 2:
        raise ValueError("Only one of (output_file, output_fp, output_file_list, output_fp_list) can be set. "
                         "(You don't have to set either one.)")
    if options is not None and options_file is not None:
        raise ValueError("Only one of options and options_file can be set. (You don't have to set either one.)")
    if multiplicity is not None and not (is_integer(multiplicity) and 0 <= multiplicity):
        raise ValueError(f"Illegal argument 'multiplicity': {multiplicity}")

    # スキーマファイルを読み込み
    if schema_file is not None:
        schema = schemas.load(schema_file)

    # スキーマの不正判定
    if not schema_is_validated:
        schemas.validate(schema)

    # オプションファイルを読み込み
    if options_file is not None:
        options = load_options(options_file)

    # ファイル出力先を正規化
    if output_file is not None:
        output_list = ((output_file, None),)
    elif output_fp is not None:
        output_list = ((None, output_fp),)
    elif output_file_list is not None:
        output_list = itertools.zip_longest(output_file_list, tuple())
    elif output_fp_list is not None:
        output_list = itertools.zip_longest(tuple(), output_fp_list)
    else:
        output_list = ((None, None),)

    factory = ranjg.Factory(schema, schema_is_validated=True)

    # メソッドの戻り値を保持するリストを作成
    if return_none:
        # 戻り値による出力を行わない場合、要素を保持しないダミーリストを使用する。
        # これにより、生成したものがリスト内に残らなくなり、早期にガベージコレクションされるようになる。
        result_list = _DummyList()
    else:
        # 戻り値を返す場合、生成したものをすべて保持する必要があるため、リストに保持する。
        result_list = []

    # 出力先の数だけ生成処理を繰り返す。
    for output_file, output_fp in output_list:

        # ランダムに値を生成
        if multiplicity is None:
            generated = factory.gen(options=options, context=context)
            result_list.append(generated)
        else:
            generated = [factory.gen(options=options, context=context) for _ in range(multiplicity)]
            result_list.extend(generated)

        # 出力先指定がある場合、JSONとして出力する
        if output_file is not None:
            with open(output_file, "w+") as fp:
                json.dump(generated, fp)
        if output_fp is not None:
            json.dump(generated, output_fp)

    if return_none:
        return None
    elif output_file_list is None and output_fp_list is None and multiplicity is None:
        return result_list[0]
    else:
        return result_list


class _DummyList(list):
    """List to ignore change operations.
    """

    def __setitem__(self, key, value):
        # ignore change operations
        pass

    def append(self, __object) -> None:
        # ignore change operations
        pass

    def extend(self, __iterable) -> None:
        # ignore change operations
        pass

    def __getitem__(self, key):
        return None

    def __len__(self) -> int:
        return 0



from setuptools import setup, find_packages


def load_readme() -> str:
    with open("README.md", encoding="utf-8") as f:
        return f.read()


setup(
    name='ranjg',
    version='0.5.1.0',
    project_urls={
        "Bug Tracker": 'https://github.com/unaguna/random-json-generator/issues',
        "Documentation": 'https://unaguna.github.io/random-json-generator/',
        "Source Code": 'https://github.com/unaguna/random-json-generator',
    },
    author='k-izumi',
    author_email='k.izumi.ysk@gmail.com',
    maintainer='k-izumi',
    maintainer_email='k.izumi.ysk@gmail.com',
    description='Generate json text randomly',
    long_description=load_readme(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=('ranjg.test',)),
    install_requires=[
        'jsonschema',
        'rstr',
    ],
    license="MIT",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Topic :: Utilities',
    ],
)
